from pathlib import Path

import pytest

from nbstore.notebook import Notebook
from nbstore.store import Store


@pytest.fixture(scope="module")
def nb(store: Store):
    return store.get_notebook("add.ipynb")


def test_add_delete(nb: Notebook):
    nb.add_data("add", "text/plain", "text")
    data = nb.get_data("add")
    assert data["text/plain"] == "text"
    nb.delete_data("add", "text/plain")
    data = nb.get_data("add")
    assert "text/plain" not in data


def test_language(nb: Notebook):
    assert nb.get_language() == "python"


def test_write(nb: Notebook, tmp_path: Path):
    path = tmp_path / "tmp.ipynb"
    nb.write(path)
    assert path.exists()
    nb2 = Notebook(path)
    data = nb2.get_mime_content("add")
    assert data
    assert data[0] == "image/png"


def test_cell_error(nb: Notebook):
    with pytest.raises(ValueError, match="Unknown identifier: unknown"):
        nb.get_cell("unknown")


def test_stream_none(nb: Notebook):
    assert nb.get_stream("add") is None


def test_data_empty(nb: Notebook):
    assert nb.get_data("empty") == {}


def test_source_include_identifier(nb: Notebook):
    source = nb.get_source("add", include_identifier=True)
    assert source.startswith("# #add\n")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from .notebook import Notebook

if TYPE_CHECKING:
    from collections.abc import Iterable


class Store:
    src_dirs: list[Path]
    active_path: Path | None
    notebooks: dict[Path, Notebook]
    st_mtime: dict[Path, float]

    def __init__(self, src_dirs: Path | str | Iterable[Path | str]) -> None:
        if isinstance(src_dirs, (str, Path)):
            src_dirs = [src_dirs]

        self.src_dirs = [Path(src_dir) for src_dir in src_dirs]
        self.active_path = None
        self.notebooks = {}
        self.st_mtime = {}

    def find_path(self, url: str) -> Path:
        if not url:
            if self.active_path:
                return self.active_path

            msg = "No active path."
            raise ValueError(msg)

        for src_dir in self.src_dirs:
            abs_path = (src_dir / url).absolute()
            if abs_path.exists():
                self.active_path = abs_path
                return abs_path

        msg = f"Source file not found in any source directory: {url}"
        raise ValueError(msg)

    def is_dirty(self, url: str) -> bool:
        path = self.find_path(url)
        if path not in self.st_mtime:
            return True

        return self.st_mtime[path] != path.stat().st_mtime

    def get_notebook(self, url: str) -> Notebook:
        path = self.find_path(url)
        st_mtime = path.stat().st_mtime

        if self.st_mtime.get(path) != st_mtime:
            self.st_mtime[path] = st_mtime
            self.notebooks[path] = Notebook(path)

        return self.notebooks[path]

    def needs_execution(self, url: str) -> bool:
        notebook = self.get_notebook(url)
        return not notebook.is_executed



import time
from pathlib import Path

import pytest

from nbstore.store import Store


def test_find_path(store: Store):
    path = store.find_path("add.ipynb")
    assert path.name == "add.ipynb"
    path = store.find_path("")
    assert path.name == "add.ipynb"


def test_find_path_error(store: Store):
    store.active_path = None
    with pytest.raises(ValueError, match="No active path."):
        store.find_path("")


def test_find_path_error_not_found(store: Store):
    with pytest.raises(ValueError, match="Source file not found"):
        store.find_path("unknown")


def test_write(store: Store, tmp_path: Path):
    nb = store.get_notebook("add.ipynb")
    path = tmp_path / "tmp.ipynb"
    nb.write(path)
    store = Store(tmp_path)
    assert store.is_dirty("tmp.ipynb")
    nb = store.get_notebook("tmp.ipynb")
    assert not store.is_dirty("tmp.ipynb")
    time.sleep(0.1)
    nb.write()
    assert store.is_dirty("tmp.ipynb")


def test_needs_execution(store: Store):
    assert store.needs_execution("add.ipynb")
    nb = store.get_notebook("add.ipynb")
    nb.execute()
    assert not store.needs_execution("add.ipynb")



import base64

import pytest

from nbstore.notebook import Notebook
from nbstore.store import Store


@pytest.fixture(scope="module", params=["holoviews", "hvplot"])
def lib(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture(scope="module")
def nb(store: Store, lib: str):
    nb = store.get_notebook(f"{lib}.ipynb")
    assert not nb.is_executed
    nb.execute()
    assert nb.is_executed
    return nb


@pytest.fixture(scope="module")
def data(nb: Notebook, lib: str):
    return nb.get_data(f"fig:{lib}")


def test_type(data: dict):
    assert isinstance(data, dict)


def test_len(data: dict):
    assert len(data) == 3


@pytest.mark.parametrize("mime", ["text/html", "text/pgf", "text/plain"])
def test_mime(data: dict, mime):
    assert mime in data


def test_data_decode(data: dict):
    text = data["text/pgf"]
    assert isinstance(text, str)
    text = base64.b64decode(text).decode(encoding="utf-8")
    assert text.startswith("%% Creator: Matplotlib,")



import pytest

from nbstore.notebook import Notebook
from nbstore.store import Store


@pytest.fixture(scope="module")
def nb(store: Store):
    nb = store.get_notebook("pgf.ipynb")
    assert not nb.is_executed
    nb.execute()
    assert nb.is_executed
    return nb


def test_cell(nb: Notebook):
    cell = nb.get_cell("fig:pgf")
    assert isinstance(cell, dict)
    assert "cell_type" in cell


def test_source(nb: Notebook):
    source = nb.get_source("fig:pgf")
    assert isinstance(source, str)
    assert source.startswith("import")
    assert "plot" in source


def test_outputs(nb: Notebook):
    outputs = nb.get_outputs("fig:pgf")
    assert isinstance(outputs, list)
    assert len(outputs) == 1
    assert isinstance(outputs[0], dict)
    assert outputs[0]["output_type"] == "display_data"
    assert "text/plain" in outputs[0]["data"]


def test_data(nb: Notebook):
    data = nb.get_data("fig:pgf")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data
    assert "image/png" in data
    assert data["text/plain"].startswith("%% Creator: Matplotlib,")


def test_stream(nb: Notebook):
    assert nb.get_stream("fig:stream") == "123\n"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import nbformat

import nbstore.pgf

from .content import get_mime_content

if TYPE_CHECKING:
    from typing import Self

    from nbformat import NotebookNode


class Notebook:
    path: Path
    node: NotebookNode
    is_executed: bool

    def __init__(self, path: str | Path) -> None:
        self.path = Path(path)
        self.node = nbformat.read(self.path, as_version=4)  # type: ignore
        self.is_executed = False

    def write(self, path: str | Path | None = None) -> None:
        nbformat.write(self.node, path or self.path)

    def get_cell(self, identifier: str) -> dict[str, Any]:
        return get_cell(self.node, identifier)

    def get_source(
        self,
        identifier: str,
        *,
        include_identifier: bool = False,
    ) -> str:
        return get_source(self.node, identifier, include_identifier=include_identifier)

    def get_outputs(self, identifier: str) -> list:
        return get_outputs(self.node, identifier)

    def get_stream(self, identifier: str) -> str | None:
        outputs = self.get_outputs(identifier)
        return get_stream(outputs)

    def get_data(self, identifier: str) -> dict[str, str]:
        outputs = self.get_outputs(identifier)
        data = get_data(outputs)
        return convert(data)

    def add_data(self, identifier: str, mime: str, data: str) -> None:
        outputs = self.get_outputs(identifier)
        if output := get_data_by_type(outputs, "display_data"):
            output[mime] = data

    def delete_data(self, identifier: str, mime: str) -> None:
        outputs = self.get_outputs(identifier)
        output = get_data_by_type(outputs, "display_data")
        if output and mime in output:
            del output[mime]

    def get_language(self) -> str:
        return get_language(self.node)

    def execute(self, timeout: int = 600) -> Self:
        try:
            from nbconvert.preprocessors import ExecutePreprocessor
        except ModuleNotFoundError:  # no cov
            msg = "nbconvert is not installed"
            raise ModuleNotFoundError(msg) from None

        ep = ExecutePreprocessor(timeout=timeout)
        ep.preprocess(self.node)
        self.is_executed = True
        return self

    def get_mime_content(self, identifier: str) -> tuple[str, str | bytes] | None:
        data = self.get_data(identifier)
        return get_mime_content(data)


def get_cell(node: NotebookNode, identifier: str) -> dict[str, Any]:
    for cell in node["cells"]:
        source: str = cell["source"]
        if source.startswith(f"# #{identifier}\n"):
            return cell

    msg = f"Unknown identifier: {identifier}"
    raise ValueError(msg)


def get_source(
    node: NotebookNode,
    identifier: str,
    *,
    include_identifier: bool = False,
) -> str:
    if source := get_cell(node, identifier).get("source", ""):
        if include_identifier:
            return source

        return source.split("\n", 1)[1]

    raise NotImplementedError


def get_outputs(node: NotebookNode, identifier: str) -> list:
    return get_cell(node, identifier).get("outputs", [])


def get_data_by_type(outputs: list, output_type: str) -> dict[str, str] | None:
    for output in outputs:
        if output["output_type"] == output_type:
            if output_type == "stream":
                return {"text/plain": output["text"]}

            return output["data"]

    return None


def get_stream(outputs: list) -> str | None:
    if data := get_data_by_type(outputs, "stream"):
        return data["text/plain"]

    return None


def get_data(outputs: list) -> dict[str, str]:
    for type_ in ["display_data", "execute_result", "stream"]:
        if data := get_data_by_type(outputs, type_):
            return data

    return {}


def get_language(node: NotebookNode) -> str:
    return node["metadata"]["kernelspec"]["language"]


def convert(data: dict[str, str]) -> dict[str, str]:
    text = data.get("text/plain")
    if text and text.startswith("%% Creator: Matplotlib, PGF backend"):
        data["text/plain"] = nbstore.pgf.convert(text)

    return data



import pytest

from nbstore.notebook import Notebook
from nbstore.store import Store


@pytest.fixture(scope="module")
def nb(store: Store):
    nb = store.get_notebook("text.ipynb")
    assert not nb.is_executed
    nb.execute()
    assert nb.is_executed
    return nb


def test_mime_content_stdout(nb: Notebook):
    content = nb.get_mime_content("text:stdout")
    assert isinstance(content, tuple)
    assert content[0] == "text/plain"
    assert content[1] == "'stdout'"


def test_mime_content_stream(nb: Notebook):
    content = nb.get_mime_content("text:stream")
    assert isinstance(content, tuple)
    assert content[0] == "text/plain"
    assert content[1] == "stream1\nstream2\n"


def test_mime_content_both(nb: Notebook):
    content = nb.get_mime_content("text:both")
    assert isinstance(content, tuple)
    assert content[0] == "text/plain"
    assert content[1] == "'hello'"


def test_mime_content_pandas(nb: Notebook):
    content = nb.get_mime_content("text:pandas")
    assert isinstance(content, tuple)
    assert content[0] == "text/html"
    assert isinstance(content[1], str)
    assert content[1].startswith("<div>")


def test_mime_content_polars(nb: Notebook):
    content = nb.get_mime_content("text:polars")
    assert isinstance(content, tuple)
    assert content[0] == "text/html"
    assert isinstance(content[1], str)
    assert content[1].startswith("<div>")
